--------------------------------------------------------------------------
-- Package cleo
-- Com tipos bsicos e codificao de instrues
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

package cleo is
    constant bus_size: integer :=8;
    subtype opcode  is std_logic_vector(2 downto 0);
    subtype internal_bus is std_logic_vector(bus_size-1 downto 0);
    
    type microinstrucao is record 
               w,r,u: opcode;
               ce, rw, lnz, lcv : std_logic;
    end record;
      
end cleo;


-------------------------------------------------------------------------
-- Decodificador de Escrita
-- Este bloco combinacional gera sinais de habilitao de escrita nos 
-- registradores do Bloco de Dados.
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity write_decoder is
    port(write_reg :in std_logic_vector(2 downto 0); hold :in std_logic;
        w_mar, w_mdr, w_ir, w_pc, w_ac, w_rs :out std_logic);
end write_decoder;

architecture write_decoder of write_decoder is
begin
    ---
    --- Cdigos de escrita nos registradores (codificao de escrita):
    --- 0 MAR, 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 PC/MDR, 7 nenhum reg. alterado
    ---
    w_mar <= '1' when hold='0' and write_reg="000"                          else '0';
    w_mdr <= '1' when hold='0' and (write_reg="001" or write_reg="110") else '0';
    w_ir  <= '1' when hold='0' and write_reg="010"                              else '0';
    w_pc  <= '1' when hold='0' and (write_reg="011" or write_reg="110")         else '0';
    w_ac  <= '1' when hold='0' and write_reg="100"                              else '0';
    w_rs  <= '1' when hold='0' and write_reg="101"                              else '0';    
end write_decoder;                                                                                                                                       


-------------------------------------------------------------------------
-- Decodificador de Leitura
-- Este bloco combinacional gera sinais de habilitao de dispositivos
-- tristate interpostos entre os registradores e barramentos.
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

entity read_decoder is
    port(read_reg:in std_logic_vector(2 downto 0);
        r_mdr, r_ir, r_pc, r_ac, r_rs: out std_logic);
end read_decoder;

architecture read_decoder of read_decoder is
begin
    ---
    --- Cdigos de leitura nos registradores (codificao de leitura):
    --- 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 MDR/AC, 7 MDR/PC
    ---
    r_mdr <= '1' when read_reg="001" or read_reg="110" or read_reg="111" else '0';
    r_ir  <= '1' when read_reg="010"                                    else '0'; 
    r_pc  <= '1' when read_reg="011" or read_reg="111"                  else '0';
    r_ac  <= '1' when read_reg="100" or read_reg="110"                  else '0';  
    r_rs  <= '1' when read_reg="101"                                    else '0';
end read_decoder; 


-------------------------------------------------------------------------
-- Unidade lgico-aritmtica (ALU)
-- A ALU possui duas entradas de dados (A_bus e B_bus) e 
-- uma saida de dados (C_bus), um sinal de controle de operao a executar
-- (alu_op) e quatro sadas com valores instantneos computados para os
-- 4 qualificadores de estado (n_inst, z_inst, c_inst, v_inst).
-------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity alu is
    port(A_bus, B_bus:  in std_logic_vector(7 downto 0);
         alu_op:        in std_logic_vector(2 downto 0);
         C_Bus:         out std_logic_vector(7 downto 0); 
         n_inst, z_inst, c_inst, v_inst: out std_logic);
end alu; 

architecture alu of alu is

   signal A_bus_int, B_bus_int, C_bus_int : std_logic_vector(8 downto 0);

begin

    A_bus_int <= '0' & A_bus;  
        
    B_bus_int <= '0' & B_bus;
        
    C_bus <= C_bus_int (7 downto 0); -- toma n-1 bits menos significativos

     process(alu_op, A_bus_int, B_bus_int)
      begin
       case alu_op is
        when "000" =>  C_bus_int <= A_bus_int + B_bus_int;       -- soma
        when "001" =>  C_bus_int <= A_bus_int + 1;               -- inc
        when "010" =>  C_bus_int <= not A_bus_int;               -- not
        when "100" =>  C_bus_int <= B_bus_int;                   -- passa B
        when "101" =>  C_bus_int <= A_bus_int or B_bus_int;      -- ou lgico
        when "110" =>  C_bus_int <= A_bus_int and B_bus_int;     -- e lgico
        when "111" =>  C_bus_int <= A_bus_int;                   -- passa A
        when others => C_bus_int <= A_bus_int;                   -- default: passa A;
       end case;
     end process;                                         
     
     n_inst <= C_bus_int(7);
     
     z_inst <= '1' when (C_bus_int(7 downto 0) = "00000000")    else '0';

     c_inst <= C_bus_int(8);

     v_inst <= '1' when (A_bus_int(7)=B_bus_int(7) and 
                         A_bus_int(7)/=C_bus_int(7)) else '0';
    
end alu;


-------------------------------------------------------------------------
-- Biestveis de estado
-- Este registrador de estado armazena quatro bits referentes aos 4
-- valores instantneos dos qualificadores de estado
-- (n_inst, z_inst, c_inst, v_inst) quando os sinais de controle 
-- (lnz e lcv) indicam que isto deva ocorrer, na borda correta do sinal de
-- relgio (clock). O sinal reset inicializa todos os qualificadores com 0.
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity status_flags is
    port(n_inst, z_inst, c_inst, v_inst, lnz, lcv, clock, reset:in std_logic;
        n, z, c, v: out std_logic);
end status_flags;

architecture status_flags of status_flags is
begin
    process(clock,reset)
    begin
        if (reset='1') then
          n <= '0';
          z <= '0';
          c <= '0';    
          v <= '0';    
        elsif clock'event and clock='0' then
          if lcv='1' then
                                        c <= c_inst;
                                        v <= v_inst;
          end if;
          if lnz='1' then
                                        n <= n_inst;
                                        z <= z_inst;
          end if;
        end if;
    end process;                   
end status_flags;


-------------------------------------------------------------------------
-- Registrador Bsico
-- Com reset assncrono e habilitao de escrita
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

entity reg_clear is
    port(clock, reset, ce:in std_logic; 
          D:in std_logic_vector(7 downto 0); 
          Q:out std_logic_vector(7 downto 0)
        ) ;
end reg_clear;

architecture reg_clear of reg_clear is
begin

  process (clock, reset)
    begin
    if reset='1' then
         Q <= (others=>'0');
    elsif (clock'event and clock='0')then    -- SENSIBILIDADE  BORDA DE DESCIDA
        if ce='1' then  
            Q <= D;  
        end if;
    end if;
  end process;
  
end reg_clear;


--########################################################################
--
--  BLOCO DE DADOS COMPLETO DA CPU
--
--########################################################################
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity datapath is
    port(clock, reset, hold     : in std_logic;
         address, ir    : out internal_bus;
         datain                 : in internal_bus;
         dataout                : out internal_bus;
         uins                   : in microinstrucao;
         n, z, c, v             : out std_logic
        ); 
end datapath;

architecture datapath of datapath is
signal  r_mdr,r_ir,r_pc,r_ac,r_rs,
                w_mar,w_mdr,w_ir,w_pc,w_ac,w_rs,
                n_inst, z_inst, c_inst, v_inst : std_logic; 
signal  outmux, mdr, ir_int, pc, ac, rs,
                busA, busB, busC : internal_bus;
 
begin  

    ir <= ir_int;    -- instrucao corrente, a ser utilizada no bloco de controle
        
        --
        -- barramento de sada do processador  sempre a sada a ALU
        --
    dataout <= busC;
    
        --
        -- cria multiplexador que gera entrada para o registrador MDR e seu controle
        --
    outmux <= busC  when (uins.ce and uins.rw)='0' else datain;
                
    --
    -- cria registradores de dados (mar, mdr, ir, pc, ac, rs),
        -- instancia-os e conecta-os 
    --
    REG_MAR: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_mar, D=>busC, Q=>address);
    REG_MDR: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_mdr, D=>outmux,  Q=>mdr);
    REG_IR: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_ir,  D=>busC, Q=>ir_int);
    REG_PC: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_pc,  D=>busC, Q=>pc);
    REG_AC: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_ac,  D=>busC, Q=>ac);
    REG_RS: entity work.reg_clear port map (clock=>clock, reset=>reset,
                ce=>w_rs,  D=>busC, Q=>rs);
    
        --
        -- instancia decodificador de escrita e conecta com habilitaes de
        -- escrita dos registradores
        --
    W_DECOD: entity work.write_decoder port map(write_reg=>uins.w, 
                hold=>hold,
                w_mar=>w_mar, w_mdr=>w_mdr, w_ir=>w_ir,
                w_pc=>w_pc, w_ac=>w_ac, w_rs=>w_rs);
    
        --
        -- instancia decodificador de leitura e conecta com habilitaes de
        -- tristates que controlam o que vai para barramentos busA e busB
        --
    R_DECOD: entity work.read_decoder port map(read_reg=>uins.r,
                r_mdr=>r_mdr, r_ir=>r_ir, r_pc=>r_pc, r_ac=>r_ac, r_rs=>r_rs);
                
    --
    -- cria tristates para jogar o valor de 1 ou 2 registradores
        -- nos barramentos de entrada da ALU, busA e busB
    --
    busB <= mdr     when r_mdr='1' else (others=>'Z');
    busB <= ir_int  when r_ir='1'  else (others=>'Z');
    busA <= pc      when r_pc='1'  else (others=>'Z');
    busA <= ac      when r_ac='1'  else (others=>'Z');
    busA <= rs      when r_rs='1'  else (others=>'Z');
    
        --
        -- instancia ALU e a conecta com barramentos busA, busB e busC,
        -- bem como com a entrada de controle e o registrador de estado
        --
    ALU: entity work.alu port map(A_bus=>busA, B_bus=>busB, C_bus=>busC,
                alu_op=>uins.u, n_inst=>n_inst, z_inst=>z_inst,
                c_inst=>c_inst, v_inst=>v_inst);

        --
        -- instancia o registrador de estado e conecta-o  ALU e a pinos da 
        -- interface com o bloco de controle
        --
    REG_STATUS: entity work.status_flags port map(n_inst=>n_inst, 
                z_inst=>z_inst, c_inst=>c_inst, v_inst=>v_inst, lnz=>uins.lnz,
                lcv=>uins.lcv, clock=>clock, reset=>reset, n=>n, z=>z,
                c=>c, v=>v);

end datapath;

